// <clown-build> csc.exe /debug+ /r:Gregor.Core.dll,Gregor.NetConsole.dll DocumentAssembly.cs </clown-build>

#define DEBUG

using System;
using System.Collections;
using System.Globalization;
using System.IO;
using System.Reflection;
using System.Text;
using System.Xml;
using Gregor.Core;
using Gregor.Core.Collections;
using Gregor.Core.Configuration;
using Gregor.NetConsole;
using Gregor.NetConsole.Engine;

[Module()]
public class Start {

    public static void Main(string[] args){
        Dev.TraceCallback = Dev.ConsoleTraceCallback;
        CCommandLine cmdLine = new CCommandLine(args);
        if(false == cmdLine.HasOptionWithAnyValue("/pathinfo", true) || cmdLine.OperandCount < 1){
            Dev.Trace("Syntax:");
            Dev.Trace("    DocumentAssembly /pathinfo:<path-info-file>[+] <assembly-name>[+]");
            Dev.Trace("Where:");
            Dev.Trace("    path-info-file ::= file name of XML file with path information for graph");
            Dev.Trace("    assembly-name  ::= name DLL");
            Dev.Trace("Example:");
            Dev.Trace("    DocumentAssembly /pathinfo:MyPathInfo.xml /pathinfo:TocPathInfo mscorlib.dll My.dll");
            Environment.Exit(1);
        }
        try{
            Run(cmdLine);
        }catch(Exception ex){
            Dev.ProcessException(ex);
        }
    }

    private static void Run(CCommandLine cmdLine){
        // create helpers
        CRoot root = new CRoot();
        CCurlyInterpreter interpreter = new CCurlyInterpreter();
        interpreter.ImportNamespace("Gregor.Core");
        CDocumentAssemblyContext context = new CDocumentAssemblyContext();
        interpreter.AddObject("Context", context); // ... provide fallback in CObjectGraph.CreateWithPathInfo
        // load assemblies
        CAssemblyLoader loader = new CAssemblyLoader();
        foreach(string sDll in cmdLine.GetOperands()){
            Assembly a = loader.LoadAssembly(sDll);
            CAssemblyInfoEx assyInfo = new CAssemblyInfoEx(a);
            root.Assemblies.Add(assyInfo);
        }
        // process
        foreach(string sOption in cmdLine.GetOptions()){
            string sPathInfoFile = Parse.ExtractAfter(sOption, ":"); // ... improve command-line API
            XmlDocument xdPathInfo = Dev.ReadXmlFile(sPathInfoFile);
            Process(root, xdPathInfo, interpreter);
        }
        // done
        Dev.Trace("Done.");
    }

    private static void Process(object root, XmlDocument xdPathInfo, ICodeInterpreter2 interpreter){
        // setup text generation info
        CTextGenerationInfo info = new CTextGenerationInfo(root);
        info.Interpreter = interpreter;
        info.InputDirectory = Environment.CurrentDirectory;
        info.OutputDirectory = Environment.CurrentDirectory;
        info.PathInfoDocument = xdPathInfo;
        // process
        CTextGenerator generator = new CTextGenerator();
        generator.Generate(info);
    }

} // module Start

public class CRoot {

    private IList m_Assemblies;

    public CRoot(){
        m_Assemblies = new ArrayList();
    }

    public IList Assemblies{
        get{
            return m_Assemblies;
        }
    }

} // class CRoot

public class CDocumentAssemblyContext : CGraphContext {

    private char[]          m_BadNameChars = {'_', '<', '>', '$'};
    private ISyntaxProvider m_SyntaxProvider;
    private CCodeOptions    m_CodeOptions;

    public CDocumentAssemblyContext() : base() {
        m_SyntaxProvider = new CCSharpSyntaxProvider();
        m_CodeOptions = new CCodeOptions();
        m_CodeOptions.UseAliasNames = false;
        m_CodeOptions.UseFullNames = true;
        m_CodeOptions.UseDefaultModifiers = false;
        m_CodeOptions.UseUnderlyingValue = true;
        m_CodeOptions.HideEnclosingTypes = false;
        m_CodeOptions.HideTypeParameters = true;
        m_CodeOptions.DecomposeValues = false;
        m_CodeOptions.EncodeChars = true;
    }

    public ISyntaxProvider SyntaxProvider{
        get{
            return m_SyntaxProvider;
        }
    }

    public CCodeOptions CodeOptions{
        get{
            return m_CodeOptions;
        }
    }

    public override string ToString(){
        return "DocumentAssemblyContext";
    }

    public string GetDateText(){
        return DateTime.Now.ToString("dddd, MMMM d, yyyy", CultureInfo.InvariantCulture);
    }

    public string GetTypeNameForFile(CTypeInfoEx typeInfo){
        Type tp = typeInfo.InnerType;
        // List`1[System.Int32]
        string s = tp.ToString();
        int iBracket = s.IndexOf('[');
        if(iBracket > 0){ // yes, > 0
            s = s.Substring(0, iBracket);
        }
        s = s.Replace(Chars.BACKTICK, '_');
        s = s.Replace('+', '.');
        return s;
    }

    public bool FilterType(CTypeInfoEx typeInfo){
        return (typeInfo.FullName.IndexOfAny(m_BadNameChars) < 0);
    }

    public ArrayList GetRelatedTypes(CTypeInfoEx typeInfo){
        Type[] aTypes = typeInfo.GetRelatedTypes(TypeRelationOptions.Default | TypeRelationOptions.BaseTypes);
        ArrayList types = WrapTypes(aTypes);
        foreach(CMemberInfoEx memberInfo in typeInfo.GetMembers()){
            foreach(CTypeInfoEx relTypeInfo in GetRelatedTypes(memberInfo)){
                if(false == types.Contains(relTypeInfo)){ // ... won't work
                    types.Add(relTypeInfo);
                }
            }
        }
        SortTypes(types);
        return types;
    }

    public ArrayList GetRelatedTypes(CMemberInfoEx memberInfo){
        Type[] aTypes = memberInfo.GetRelatedTypes(TypeRelationOptions.Default);
        ArrayList types = WrapTypes(aTypes);
        SortTypes(types);
        return types;
    }

    public string GetTypeDetails(CTypeInfoEx typeInfo){
    	StringBuilder sb = new StringBuilder();
    	// signature
        sb.Append(typeInfo.GetSignature(m_SyntaxProvider, m_CodeOptions));
        sb.Append(Environment.NewLine);
        sb.Append(Environment.NewLine);
        // bases
        string sBases = typeInfo.GetBasesDeclarationString(m_SyntaxProvider, m_CodeOptions);
        if(Flow.IsString(sBases)){
        	sb.Append(sBases);
        	sb.Append(Environment.NewLine);
        	sb.Append(Environment.NewLine);
        }
        // custom attributes
        sb.Append("Custom Attributes: ");
        sb.Append(typeInfo.GetCustomAttributesString(m_SyntaxProvider, m_CodeOptions));
        sb.Append(Environment.NewLine);
        sb.Append(Environment.NewLine);
        // description
        sb.Append("Description: ");
        sb.Append(typeInfo.GetDescription());
        sb.Append(Environment.NewLine);
        sb.Append(Environment.NewLine);
        // nick name
        string sAlias = typeInfo.GetAliasName(m_SyntaxProvider);
        if(Flow.IsString(sAlias)){
        	sb.Append(m_SyntaxProvider.LanguageName + " Alias Name: ");
        	sb.Append(sAlias);
        	sb.Append(Environment.NewLine);
        	sb.Append(Environment.NewLine);
        }
        // done
        return sb.ToString();
    }

    public string GetMemberDetails(CMemberInfoEx memberInfo){
        StringBuilder sb = new StringBuilder();
        // signature
        sb.Append(memberInfo.GetSignature(m_SyntaxProvider, m_CodeOptions).Trim());
        // interface mapping clause or comment
        sb.Append(memberInfo.GetInterfaceMappingString(m_SyntaxProvider, m_CodeOptions));
        sb.Append(Environment.NewLine);
        sb.Append(Environment.NewLine);
        // attributes
        sb.Append("Custom Attributes: ");
        sb.Append(memberInfo.GetCustomAttributesString(m_SyntaxProvider, m_CodeOptions));
        sb.Append(Environment.NewLine);
        sb.Append(GetParameterCustomAttributeDetails(memberInfo));
        // description
        sb.Append("Description: ");
        sb.Append(memberInfo.GetDescription());
        sb.Append(Environment.NewLine);
        sb.Append(GetParameterDescriptionDetails(memberInfo));
        // extra init info
        CCodeOptions options = m_CodeOptions.Copy();
        options.UseUnderlyingValue = false;
        options.DecomposeValues = true;
        string sExtra = memberInfo.GetExtraInfo("InitInfo", m_SyntaxProvider, options);
        if(Flow.IsString(sExtra)){
            sb.Append("Value (Decomposed): ");
            sb.Append(sExtra);
            sb.Append(Environment.NewLine);
            sb.Append(Environment.NewLine);
        }
        // done
        return sb.ToString();
    }

    public string GetParameterCustomAttributeDetails(CMemberInfoEx memberInfo){
        StringBuilder sb = new StringBuilder();
        CParameterInfoEx[] paramInfos = memberInfo.GetParameters();
        if(paramInfos != null){
            foreach(CParameterInfoEx paramInfo in paramInfos){
                string s = paramInfo.GetCustomAttributesString(m_SyntaxProvider, m_CodeOptions);
                if(Flow.IsString(s)){
    				sb.Append(paramInfo.Name);
    				sb.Append(" - ");
    				sb.Append(s);
    				sb.Append(Environment.NewLine);
                }
            }    
        }
        sb.Append(Environment.NewLine);
        return sb.ToString();
    }
    
    public string GetParameterDescriptionDetails(CMemberInfoEx memberInfo){
        StringBuilder sb = new StringBuilder();
        CParameterInfoEx[] paramInfos = memberInfo.GetParameters();
        if(paramInfos != null){
            foreach(CParameterInfoEx paramInfo in paramInfos){
                string s = paramInfo.GetDescription();
                if(Flow.IsString(s)){
    				sb.Append(paramInfo.Name);
    				sb.Append(" - ");
    				sb.Append(s);
    				sb.Append(Environment.NewLine);
                }
            }
        }
        return sb.ToString();
    }

    private ArrayList WrapTypes(Type[] aTypes){
        ArrayList ret = new ArrayList(aTypes.Length);
        foreach(Type tp in aTypes){
            ret.Add(new CTypeInfoEx(tp));
        }
        return ret;
    }

    private void SortTypes(ArrayList types){
        CTypeInfoExComparer comparer = new CTypeInfoExComparer();
        comparer.GroupByNamespace = true;
        comparer.GroupByTypeKind = true;
        types.Sort(comparer);
    }

} // class CDocumentAssemblyContext
