// <clown-build>csc.exe /debug+ /r:Gregor.Core.dll ListProjectOrphans.cs</clown-build>
// <clown-run>ListProjectOrphans.exe</clown-run>

using System;
using System.IO;
using System.Xml;
using Gregor.Core;

namespace Gregor {

[Module()]
public class ListProjectOrphans {

    public static void Main(string[] args) {

        if(args.Length != 2
        || false == Flow.IsAmong(args[0], "2002", "2003", "2005")){
            SyntaxExit();
        }

        try{
            Run(args[0], args[1]);
        }catch(Exception ex){
            W(ex);
        }

    }

    private static void SyntaxExit(){
        W("Syntax:");
        W("    ListProjectOrphans <version> <vs-project-file> | <root-directory>");
        W("Where:");
        W("    version ::= 2002 | 2003 | 2005");
        W("Remarks:");
        W("    If the project files uses XML namespaces, this might not work;");
        W("    remove all xmlns attributes as a workarround.");
        Environment.Exit(1);
    }

    private static void Run(string sVersion, string sSpec){
        if(Directory.Exists(sSpec)){
            RunForDirectory(sVersion, sSpec);
        }else if(File.Exists(sSpec)){
            RunForSingleProject(sVersion, sSpec);
        }else{
            SyntaxExit();
        }
    }

    private static void RunForDirectory(string sVersion, string sDir){
        DirectoryInfo di = new DirectoryInfo(sDir);
        foreach(FileInfo fi in di.GetFiles()){
            string sExt = "." + Parse.ExtractAfterLast(fi.Name, ".");
            if(Flow.IsAmong(sExt, ".csproj", ".vbproj", ".vcproj")){
                try{
                    RunForSingleProject(sVersion, fi.FullName);
                }catch(Exception ex){
                    W(ex);
                }
            }
        }
        foreach(DirectoryInfo diSub in di.GetDirectories()){
            RunForDirectory(sVersion, diSub.FullName);
        }
    }

    private static void RunForSingleProject(string sVersion, string sProjectFile){

        // W("Processing project '" + sProjectFile + "' ...");
        CProjectInfo project = GetProjectInfo(sVersion, sProjectFile);

        foreach(string sExt in project.Extensions){
            foreach(FileInfo fi in project.ProjectDirectory.GetFiles("*" + sExt)){
                if(false == project.IncludesFile(fi.Name)){
                    W(fi.FullName);
                }
            }
        }
    }

    private static CProjectInfo GetProjectInfo(string sVersion, string sProjectFile){
        string sExt = "." + Parse.ExtractAfterLast(sProjectFile, ".");
        switch(sVersion){
            case "2002":
            case "2003":
                switch(sExt){
                    case ".csproj":
                        return new CProjectInfo(sProjectFile, "/VisualStudioProject/CSHARP/Files/Include/File[@RelPath='$FileName%']", ".cs");
                    case ".vbproj":
                        return new CProjectInfo(sProjectFile, "/VisualStudioProject/VisualBasic/Files/Include/File[@RelPath='$FileName%']", ".vb");
                    case ".vcproj":
                        return new CProjectInfo(sProjectFile, "/VisualStudioProject/Files/Filter/File[@RelativePath='$FileName%']", ".cpp", ".h");
                    default:
                        Check.Expression(false, "Project type '" + sExt + "' not supported.");
                        return null; // CS0161
                }
            case "2005":
                switch(sExt){
                    case ".csproj":
                        return new CProjectInfo(sProjectFile, "/Project/ItemGroup/Compile[@Include='$FileName%']", ".cs");
                    case ".vbproj":
                        return new CProjectInfo(sProjectFile, "/Project/ItemGroup/Compile[@Include='$FileName%']", ".vb");
                    case ".vcproj":
                        return new CProjectInfo(sProjectFile, "/VisualStudioProject/Files/Filter/File[@RelativePath='$FileName%']", ".cpp", ".h");
                    default:
                        Check.Expression(false, "Project type '" + sExt + "' not supported.");
                        return null; // CS0161
                }
            default:
                throw new ArgumentException(sVersion, "sVersion");
        }

    }

    internal static void W(object obj){
        Console.WriteLine(obj);
    }

} // module ListProjectOrphans

internal class CProjectInfo {

    private string m_ProjectFile;
    private XmlDocument m_ProjectDoc;
    private string m_XPathExpression;
    private string[] m_Extensions;

    public CProjectInfo(string sProjectFile, string sXPath, params string[] asExtensions){
        Check.String(sProjectFile);
        Check.String(sXPath);
        Check.Array(asExtensions, true);
        m_ProjectFile = sProjectFile;
        m_XPathExpression = sXPath;
        m_Extensions = asExtensions;
        m_ProjectDoc = new XmlDocument();
        string sXml = FileUtil.ReadTextFile(sProjectFile);
        sXml = sXml.Replace("xmlns=\"http://schemas.microsoft.com/developer/msbuild/2003\"", string.Empty);
        m_ProjectDoc.LoadXml(sXml);
    }

    public FileInfo ProjectFile{
        get{
            return new FileInfo(m_ProjectFile);
        }
    }

    public DirectoryInfo ProjectDirectory{
        get{
            return (new FileInfo(m_ProjectFile)).Directory;
        }
    }

    public XmlDocument ProjectDoc{
        get{
            return m_ProjectDoc;
        }
    }

    public string XPathExpression{
        get{
            return m_XPathExpression;
        }
    }

    public string[] Extensions{
        get{
            return m_Extensions;
        }
    }

    public bool IncludesFile(string sFileName){
        string sTmp = sFileName;
        string sXPath = m_XPathExpression.Replace("$FileName%", sTmp);
        if(this.DoesIncludeFile(sXPath)){
            return true;
        }
        sTmp = ".\\" + sFileName;
        sXPath = m_XPathExpression.Replace("$FileName%", sTmp);
        if(this.DoesIncludeFile(sXPath)){
            return true;
        }
        // ...
        return false;
    }

    private bool DoesIncludeFile(string sXPath){
        return (null != m_ProjectDoc.SelectSingleNode(sXPath));
    }

} // class CProjectInfo

} // namespace Gregor
