// <clown-build> csc.exe /debug+ /r:Gregor.Core.dll,Gregor.AppCore.dll,Gregor.Inet.dll XPath.cs </clown-build>
// <clown-run> XPath.exe </clown-run>

using System;
using System.Collections;
using System.IO;
using System.Net;
using System.Text;
using System.Xml;
using Gregor.Core;
using Gregor.Inet;

namespace Gregor {

public class Start {

    public static void Main(string[] args) {

        if(args.Length < 2){
            W("Syntax:");
            W("    XPath <xml-file> <xpath-expression> [<options>] [<attrib1> ...]");
            W("Where:");
            W("    options ::= /distinct | /sorted");
            W("Notes:");
            W("    The xml-file argument can be a relative or absolute file path or an URL.");
            W("    The attribute names can be followed by a colon and a number to indicate");
            W("    the width of the column.");
            W("Example:");
            W("    XPath http://peisker.net/journal.htm /html/body/p/a href:40");
            Environment.Exit(1);
        }

        string sUrl = args[0];
        string sXPathExpr = args[1];
        bool fDistinct = (args.Length > 2 && args[2] == "/distinct" || args.Length > 3 && args[3] == "/distinct");
        bool fSorted = (args.Length > 2 && args[2] == "/sorted" || args.Length > 3 && args[3] == "/sorted");

        try{
            string[] asAttribs = {};
            int iAttribsStart = (fDistinct && fSorted ? 4 : (fDistinct || fSorted ? 3 : 2));
            if(args.Length > iAttribsStart){
                int len = args.Length - iAttribsStart;
                asAttribs = new string[len];
                Array.Copy(args, iAttribsStart, asAttribs, 0, len);
            }
            Run(sUrl, sXPathExpr, fDistinct, fSorted, asAttribs);
        }catch(Exception ex){
            W(ex);
        }

    }

    private static void Run(string sUrl, string sXPathExpr, bool fDistinct, bool fSorted, params string[] asAttribs){
        XmlDocument xd = ReadXml(sUrl);
        IEnumerable hits = xd.SelectNodes(sXPathExpr);
        if(fDistinct || asAttribs.Length > 0){
            IEnumerable columns = GetColumns(asAttribs);
            ShowTable(hits, fDistinct, fSorted, columns);
        }else{
            ShowNodes(hits);
        }
    }

    private static XmlDocument ReadXml(string sUrl){
        XmlDocument xd = new XmlDocument();
        string sXml = NetUtil.ReadUrl(sUrl);
        xd.LoadXml(sXml);
        return xd;
    }

    private static IEnumerable GetColumns(params string[] asAttribs){
        ArrayList columns = new ArrayList();
        foreach(string sAttrib in asAttribs){
            columns.Add(new CColumnInfo(sAttrib));
        }
        return columns;
    }

    private static void ShowNodes(IEnumerable hits){
        W("<result>");
        foreach(XmlNode xn in hits){
            W(xn.OuterXml);
        }
        W("</result>");
    }

    private static void ShowTable(IEnumerable hits, bool fDistinct, bool fSorted, IEnumerable columns){
        if(fDistinct){
            hits = GetDistinctElements(hits, columns);
        }
        if(fSorted){
            hits = SortElements(hits, columns);
        }
        foreach(CColumnInfo column in columns){
            string sName = AlignString(column.Name, column.Width);
            Console.Write(sName);
        }
        W(string.Empty);
        foreach(CColumnInfo column in columns){
            string sLine = AlignString('-', column.Width);
            Console.Write(sLine);
        }
        W(string.Empty);
        int count = 0; // we can't use XmlNodeList.Count
        foreach(XmlNode xn in hits){
            foreach(CColumnInfo column in columns){
                string s = "(missing)";
                XmlAttribute xa = xn.Attributes[column.Name];
                if(xa != null){
                    s = xa.Value;
                }
                s = AlignString(s, column.Width);
                Console.Write(s);
            }
            W(string.Empty);
            count++;
        }
        foreach(CColumnInfo column in columns){
            string sLine = AlignString('-', column.Width);
            Console.Write(sLine);
        }
        W(string.Empty);
        W(count + " nodes selected.");
    }

    internal static string AlignString(char c, int width){
        return new string(c, width);
    }

    internal static string AlignString(string s, int width){
        if(s.Length >= width){
            return s.Substring(0, width - 1) + ' ';
        }else{
            return s.PadRight(width, ' ');
        }
    }

    private static IEnumerable GetDistinctElements(IEnumerable hits, IEnumerable columns){
        ArrayList list = new ArrayList();
        Hashtable table = new Hashtable();
        foreach(XmlNode xn in hits){
            StringBuilder sb = new StringBuilder(256);
            bool fAfterFirst = false;
            foreach(CColumnInfo column in columns){
                if(fAfterFirst){
                    // this isn't 100% safe
                    sb.Append("###");
                }
                XmlAttribute xa = xn.Attributes[column.Name];
                if(xa != null){
                    sb.Append(xa.Value);
                }else{
                    sb.Append(string.Empty);
                }
                fAfterFirst = true;
            }
            string sKey = sb.ToString();
            if(false == table.ContainsKey(sKey)){
                table.Add(sKey, xn);
                list.Add(xn);
            }
        }
        return list;
    }

    private static IEnumerable SortElements(IEnumerable hits, IEnumerable columns){
        ArrayList list = new ArrayList();
        foreach(XmlNode xn in hits){
            list.Add(xn);
        }
        list.Sort(new CXmlNodeComparer(columns));
        return list;
    }

    internal static void W(object obj){
        Console.WriteLine(obj);
    }

} // class Start

internal class CColumnInfo {

    private const int DEFAULT_WIDTH = 20;

    private string m_Name;
    private int m_Width;

    public CColumnInfo(string sValue){
        int iColon = sValue.IndexOf(':');
        if(iColon >= 0){
            m_Name = sValue.Substring(0, iColon);
            try{
                m_Width = int.Parse(sValue.Substring(iColon + 1));
            }catch(Exception){
            }
            if(m_Width < 1){
                m_Width = 1;
            }else if(m_Width > 150){
                m_Width = 150;
            }
        }else{
            m_Name = sValue;
            m_Width = DEFAULT_WIDTH;
        }
    }

    public string Name{
        get{
            return m_Name;
        }
    }

    public int Width{
        get{
            return m_Width;
        }
    }

} // class CColumnInfo

internal class CXmlNodeComparer : IComparer {

    private IEnumerable m_Columns;

    public CXmlNodeComparer(IEnumerable columns){
        m_Columns = columns;
    }

    public int Compare(object a, object b){
        int iRet = 0;
        XmlNode xn1 = (XmlNode) a;
        XmlNode xn2 = (XmlNode) b;
        foreach(CColumnInfo column in m_Columns){
            // allow XPath queries here
            XmlAttribute xa1 = xn1.Attributes[column.Name];
            XmlAttribute xa2 = xn2.Attributes[column.Name];
            if(xa1 != null && xa2 != null){
                iRet = string.Compare(xa1.Value, xa2.Value);
            }else if(xa1 != null && xa2 == null){
                iRet = 1;
            }else if(xa1 == null && xa2 != null){
                iRet = -1;
            }
            if(iRet != 0){
                break;
            }
        }
        return iRet;
    }

} // class CXmlNodeComparer

} // namespace Gregor
